/*
 * Raima Database Manager
 *
 * Copyright (c) 2019 Raima Inc.,  All rights reserved.
 *
 * Use of this software, whether in source code format, or in executable,
 * binary object code form, is governed by the Raima LICENSE which
 * is fully described in the LICENSE.TXT file, included within this
 * distribution of files.
 */

/*! \file
 * \brief Source code for the RDM core13 example
 */
#include <stdio.h>
#include <string.h>
#include "print_error.h"
#include "rdm.h"

/* Generated \c struct and \c typedef definitions to be used with the RDM APIs */
#include "core13_db_structs.h"

/* Generated catalog definition to be used with the RDM rdm_dbSetCatalog() API */
#include "core13_db_cat.h"

/*!
 * \mainpage Core13 Popcorn Example
 *
 * Each time you run this example:
 * \li The database is initialized (all existing contents removed)
 * \li Two RDM tasks are created
 * \li Each task opens the database is shared mode
 * \li Task 1 starts a transaction, locks the office record for writes,
 *      adds a few office records
 * \li Task 2 attempts to read the office records (fails while task 1 has lock)
 * \li Task 1 commits transaction
 * \li Task 2 is able to read the office records
 *
 * \par Table of Contents
 *
 * - \subpage hDB
 * - \subpage hRDMfunc
 * - \subpage hPGMfunc
 *
 * For additional information, please refer to the product documentation at http://docs.raima.com/.
 *
 * \page hDB Database Schema
 *
 * \par Database Schema Definition
 *
 * The DDL (Database Definition Language) specification for the database used
 * in this example is located in the file \c core13_db.sdl.
 *
 * \include core13_db.sdl
 *
 * The schema was compiled using the RDM rdm-compile utility with the -s option to
 * generate C-structures for interfacing with the database.
 *
 * \code rdm-compile -s core13_db.sdl \endcode
 *
 * \page hRDMfunc RDM Functions Used
 * The following RDM Functions are illustrated:
 * \li rdm_rdmAllocTFS/rdm_tfsFree
 * \li rdm_tfsAllocDatabase/rdm_dbFree
 * \li rdm_dbSetCatalog
 * \li rdm_dbOpen/rdm_dbClose
 * \li rdm_dbInsertRow
 * \li rdm_dbGetRowsByKey/rdm_cursorFree
 * \li rdm_cursorMoveToFirst/rdm_cursorMoveToNext
 * \li rdm_cursorReadRow
 *
 * They are fully documented in http://docs.raima.com/.
 *
 * Because the database is opened in exclusive mode lock and transactions
 * are not required.  Future core examples will use shared mode and will
 * demonstrate the use of transactions and locks.
 *
 * Each of these functions returns an integer status code, where the
 * value sOKAY indicates a successful call.
 *
 * The actual databases will be stored in a directories named 'na_sales_db' and
 * 'emea_sales_db' in the project directory.
 *
 * \page hPGMfunc Program Functions
 *
 * For simplicity, this example does not check all return codes, but good
 * programming practices would dictate that they are checked after each
 * RDM call.
 *
 * \li openEmptyDatabase() - \copybrief openEmptyDatabase
 * \li display_offices() - \copybrief display_offices
 * \li main() - \copybrief main
 */

/*! \brief Initialize the RDM runtime library for use in the core13 example
 *
 * This function initializes the RDM Transactional File Server (TFS) to use
 * the EMBEDED TFS implementation.  It also allocates a database handle
 * and opens the "core13" database in exclusive mode.
 * Exclusve mode does not require database locks or transactions.
 *
 * @return Returns RDM_RETCODE status code (sOKAY if successful)
 */

RDM_RETCODE openEmptyDatabase (RDM_TFS hTFS, /**< [in] Allocated TFS handle */
                               RDM_DB *pDB1, /**< [out] Pointer to the RDM database handle */
                               RDM_DB *pDB2) /**< [out] Pointer to the RDM database handle */
{
    RDM_RETCODE rc;

    rc = rdm_tfsAllocDatabase (hTFS, pDB1);
    print_error (rc);

    if (rc == sOKAY)
    {
        /* Override the default DURABLE setting to CONSISTENT */
        rc = rdm_dbSetOptions (*pDB1, "durability=consistent");
        print_error (rc);

        /* If we have a DB handle ... */
        rc = rdm_dbSetCatalog (*pDB1, core13_db_cat);
        print_error (rc);

        if (rc == sOKAY)
        {
            rc = rdm_dbOpen (*pDB1, "core13", RDM_OPEN_SHARED);
            print_error (rc);
        }
        if (rc == sOKAY)
        {
            rc = rdm_dbStartUpdate (*pDB1, RDM_LOCK_ALL, 0, NULL, 0, NULL);
            rc = rdm_dbDeleteAllRowsFromDatabase (*pDB1);
            print_error (rc);
            rc = rdm_dbEnd (*pDB1);
        }

        /* allocate second database handle */
        rc = rdm_tfsAllocDatabase (hTFS, pDB2);
        print_error (rc);

        /* Override the default DURABLE setting to CONSISTENT */
        rc = rdm_dbSetOptions (*pDB2, "durability=consistent");
        print_error (rc);

        if (rc == sOKAY)
        {
            rc = rdm_dbOpen (*pDB2, "core13", RDM_OPEN_SHARED);
            print_error (rc);
        }

        /* If any of the above DB operations failed, free the DB handle */
        if (rc != sOKAY)
        {
            rdm_dbFree (*pDB1);
            rdm_dbFree (*pDB2);
        }
    }
    return rc;
}

/*! \brief Display the contents of the database
 *
 * This function displays the contents of the database.
 * Exclusve mode does not require database locks or transactions.
 *
 * @return Returns RDM_RETCODE status code (sOKAY if successful)
 */
RDM_RETCODE display_offices (RDM_DB hDB) /**< [in] Database handle to be used */
{
    RDM_RETCODE rc;
    OFFICE office_rec;
    RDM_CURSOR cursor = NULL;

    rc = rdm_dbStartRead (hDB, RDM_LOCK_ALL, 0, NULL);
    print_error (rc);

    if (rc == sOKAY)
    {
        /* The following cursor association call will allocate the cursor
         *  if the cursor is set to NULL. This short-cut can eliminate the
         *  requirement to call rdm_dbAllocCursor() before using the cursor
         *  in this function */
        rc = rdm_dbGetRowsByKey (hDB, KEY_OFFICE_NAME, &cursor);
        print_error (rc);

        for (rc = rdm_cursorMoveToFirst (cursor); rc == sOKAY; rc = rdm_cursorMoveToNext (cursor))
        {
            /* Read and display the current person record */
            rc = rdm_cursorReadRow (cursor, &office_rec, sizeof (office_rec), NULL);
            print_error (rc);
            printf ("%s\n", office_rec.NAME);
        }

        /* free the cursor if it was allocated */
        if (cursor)
            rdm_cursorFree (cursor);

        /* Expect rc to be sENDOFCURSOR when we exit the loop */
        if (rc == sENDOFCURSOR)
        {
            rc = sOKAY;
        }
        rdm_dbEnd (hDB);
    }

    return rc;
}

/*! \brief Main function for core13 example
 *
 * The function initializes the RDM environment and runs the create, read operations.
 *
 * @return Returns the RDM_RETCODE on exit.
 */
int main (void)
{
    RDM_RETCODE rc;
    RDM_TFS hTFS;
    RDM_DB hDB1;
    RDM_DB hDB2;
    OFFICE office_rec;

    rc = rdm_rdmAllocTFS ("", &hTFS);

    /* create records in the NA database */
    /* Open the NA database for exclusive access, this will create the database
     * dictionary. */
    rc = openEmptyDatabase (hTFS, &hDB1, &hDB2);

    if (rc == sOKAY)
    {
        rc = rdm_dbStartUpdate (hDB1, RDM_LOCK_ALL, 0, NULL, 0, NULL);
        print_error (rc);

        if (rc == sOKAY)
        {
            strncpy (office_rec.NAME, "Seattle", sizeof (office_rec.NAME));
            rdm_dbInsertRow (hDB1, TABLE_OFFICE, &office_rec, sizeof (office_rec), NULL);

            strncpy (office_rec.NAME, "Boise", sizeof (office_rec.NAME));
            rdm_dbInsertRow (hDB1, TABLE_OFFICE, &office_rec, sizeof (office_rec), NULL);

            strncpy (office_rec.NAME, "San Francisco", sizeof (office_rec.NAME));
            rdm_dbInsertRow (hDB1, TABLE_OFFICE, &office_rec, sizeof (office_rec), NULL);

            strncpy (office_rec.NAME, "Dallas", sizeof (office_rec.NAME));
            rdm_dbInsertRow (hDB1, TABLE_OFFICE, &office_rec, sizeof (office_rec), NULL);

            /* Display all offices using task2 (should fail because we will not be
             * able to get a read lock */
            printf ("\nDisplaying Offices using database handle #2.\n");
            rc = display_offices (hDB2);

            rc = rdm_dbEnd (hDB1);
            print_error (rc);

            if (rc == sOKAY)
            {
                /* Display all offices */
                printf ("\nDisplaying Offices using database handle #2.\n");
                rc = display_offices (hDB2);
            }
        }
    }

    /* close the database */
    rdm_dbFree (hDB1);
    rdm_dbFree (hDB2);
    rdm_tfsFree (hTFS);

    return 0;
}
